﻿using MapInfo.Data;
using MapInfo.Engine;
using MapInfo.Geometry;
using MapInfo.RasterEngine.Common;
using MapInfo.RasterEngine.IO;
using MapInfo.RasterEngine.Operations;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace EnviroRisk
{
    public class TableHelper
    {

        public class PointTable
        {
            Dictionary<DPoint, double> data;

            public bool hasData()
            {
                return this.data.Keys.Count > 0;
            }

            public PointTable()
            {
                this.data = new Dictionary<DPoint, double>();
            }
            // Loads a point tab file
            public PointTable(string infile)
            {
                EnviroRiskAddIn.Debug("Loading file: " + infile);

                this.data = new Dictionary<DPoint, double>();
                var watch = System.Diagnostics.Stopwatch.StartNew();
                loadTab(infile);
                watch.Stop();
                var elapsedMs = watch.ElapsedMilliseconds;
                EnviroRiskAddIn.Debug("Load total: " + elapsedMs / 1000 + " s" + " ; " + this.data.Keys.Count + " records");
                
            }

            // Load data from tab file
            private void loadTab(string infile)
            {
                if (!File.Exists(infile))
                    return;

                TableInfo info = TableInfo.CreateFromFile(infile);
                TableSessionInfo tsi = new TableSessionInfo();
                tsi.ReadOnly = true;

                Catalog catalog = Session.Current.Catalog;
                Table table = catalog.OpenTable(infile, "TMPDATA", tsi);

                NumberFormatInfo nfi = new NumberFormatInfo();
                nfi.NumberDecimalSeparator = ".";
                this.data = new Dictionary<DPoint, double>(1038241);
                table.BeginAccess(TableAccessMode.Read);
                var reader = table.ExecuteReader();
                while (reader.Read())
                {
                    DPoint d =  reader.GetFeatureGeometry(0).Centroid;
                    double value = reader.GetDouble(2);
                    this.data.Add(d, value);
                }

                table.EndAccess();

                table.Close();
            }

            // Apply an operation on every element of the table returning a new table or modifying the first one
            public PointTable map(Func<double, double> operation, bool mutate=false)
            {
                PointTable dst = null;
                if (mutate)
                    dst = this;
                else dst = new PointTable();

                foreach (var entry in this.data)
                    dst.data[entry.Key] = operation(entry.Value);
                
                return dst;
            }

            // Apply an operation on every element of the tables returning a new table or modifying the first one
            public PointTable map(Func<double, double, double> operation, PointTable other, bool mutate=false)
            {
                PointTable dst = null;
                if (mutate)
                    dst = this;
                else dst = new PointTable();

                try
                {
                    foreach (var entry in this.data)
                        dst.data[entry.Key] = operation(entry.Value, other.data[entry.Key]);
                }
                catch (KeyNotFoundException e)
                {
                    EnviroRiskAddIn.Debug("Point Table Map exception: " + e.ToString());
                    dst.data.Clear();
                }
                return dst;
            }

            // Apply an operation on every element of the tables returning a new table or modifying the first one
            public PointTable map(Func<double, double, double, double> operation, PointTable other1, PointTable other2, bool mutate = false)
            {
                PointTable dst = null;
                if (mutate)
                    dst = this;
                else dst = new PointTable();

                try
                {
                    foreach (var entry in this.data)
                    dst.data[entry.Key] = operation(entry.Value, other1.data[entry.Key], other2.data[entry.Key]);
                }
                catch (KeyNotFoundException e)
                {
                    EnviroRiskAddIn.Debug("Point Table Map exception: " + e.ToString());
                    dst.data.Clear();
                }
                return dst; 
            }

            // write the data to an ascii file
            public void writeASCII(string filename)
            {
                NumberFormatInfo nfi = new NumberFormatInfo();
                nfi.NumberDecimalSeparator = ".";
                nfi.NumberDecimalDigits = 3;

                using (System.IO.StreamWriter fout = new StreamWriter(filename))
                {
                    foreach (var pair in this.data)
                    {
                        DPoint p = pair.Key;
                        double value = pair.Value;

                        string line = "" + p.x.ToString(nfi) +
                            " " + p.y.ToString(nfi) +
                            " " + value.ToString(nfi);
                        fout.WriteLine(line);
                    }
                }
            }


            public void writeRaster(string outfile)
            {
                string tmpfile = "tmpdata.tmp";

                var watch = System.Diagnostics.Stopwatch.StartNew();
                writeASCII(tmpfile);
                watch.Stop();
                var elapsedMs = watch.ElapsedMilliseconds;
                EnviroRiskAddIn.Debug("Write total: " + elapsedMs / 1000 + " s"); 

                var watch2 = System.Diagnostics.Stopwatch.StartNew();
                
                Interpolate(tmpfile, outfile);
                watch2.Stop();
                var elapsedMs2 = watch2.ElapsedMilliseconds;
                EnviroRiskAddIn.Debug("Interpolate total: " + elapsedMs2 / 1000 + " s");

                //File.Delete(tmpfile);
            }

            private void Interpolate(string strInputFile, string strOutputFile)
            {
                RasterInterpolationOptions interpolationOptions = new RasterInterpolationOptions
                {
                    Preferences =
                    {
                        AutoCacheSize = true,
                        CacheSize = 0,
                        TemporaryDirectory = ""
                    }
                };

                string strCoordSystem = "CoordSys Earth Projection 1, 104";
                // setup source data file parameters, can add more than one file
                RasterInterpolationInputFileDetails file = new RasterInterpolationInputFileDetails();
                file.DataFieldIndexes.Add(2);
                file.XFieldIndex = 0;
                file.YFieldIndex = 1;
                file.FileType = RasterInterpolationInputFileType.ASCII_SPACE;
                file.FilePath = strInputFile;
                file.HeaderRows = 0;
                file.CoordSysString = strCoordSystem;
                file.SubFileName = "*.*";
                interpolationOptions.InputDetails.Add(file);

                // setup output raster parameters
                interpolationOptions.OutputDetails.FilePath = strOutputFile;
                interpolationOptions.OutputDetails.DriverId = DriverID.MRR.GetString();
                interpolationOptions.OutputDetails.AutoDataType = true;

                /* not used in this example as we have set AutoDataType to true
                interpolationOptions.OutputFile.DataTypes.Add(RasterBandDataType.RealSingle);        
                */
                interpolationOptions.OutputDetails.CoordSysString = strCoordSystem;

                // setup raster geometry parameters
                interpolationOptions.Geometry.Extent.Auto = true;


                interpolationOptions.Geometry.CellSize.Auto = true;
                /* not used in this example as we have set AutoGridCellSize to true
                interpolationOptions.Geometry.CellSize.X = 10;
                interpolationOptions.Geometry.CellSize.Y = 10;
                */
                 
                // setup minimum curvature method parameters
                RasterInterpolationMinimumCurvature minimumCurvatureOptions = new RasterInterpolationMinimumCurvature();
                minimumCurvatureOptions.ParameterType.UnitsType = RasterInterpolationParameterUnitsType.Cell;
                minimumCurvatureOptions.ParameterType.UnitCode = MapInfoUnitCode.Meters;
                minimumCurvatureOptions.Iterations = RasterInterpolationIterationIntensity.Normal;

                minimumCurvatureOptions.PercentChange = 0.001;
                
                minimumCurvatureOptions.SplineTension = 0;
                minimumCurvatureOptions.StampMethod = RasterInterpolationMinimumCurvatureStampingMethod.AverageLastInWeighted;

                minimumCurvatureOptions.Smoothing.Type = RasterInterpolationSmoothingType.Gaussian;
                minimumCurvatureOptions.Smoothing.Level = 5;
                minimumCurvatureOptions.Clip.Method = RasterInterpolationClipMethod.None;
                

                RasterApiOptions apiOptions = new RasterApiOptions();

                RasterCreationOptions creationOptions = new RasterCreationOptions
                {
                    CompressionOptions = new RasterCompressionOptions(RasterCompressionType.DataBalanced,3),
                    PredictiveEncoding = EncodingType.PreviousColumnLinear
                };
                apiOptions.CreationOptions = creationOptions;

                RasterInterpolation.InterpolateMinimumCurvature(interpolationOptions, minimumCurvatureOptions, apiOptions);
            }
        }

        // Opens a table
        public static string OpenTable(string path, Form parent = null)
        {
            EnviroRiskAddIn.Debug("Opening table " + path);

            Semaphore s = new Semaphore(0,1);
            string tableName = "";
            if (parent == null)
                tableName = EnviroRiskAddIn.MBA.EvaluateMapBasicFunction("OpenTAB", path);
            else {

                parent.BeginInvoke(new Action(() =>
                {
                    try
                    {
                        
                       tableName =  EnviroRiskAddIn.MBA.EvaluateMapBasicFunction("OpenTAB", path);
                    }
                    finally
                    {
                        s.Release();
                    }
                }));
                s.WaitOne();
            }
            return tableName;
        }

        // Closes a table
        public static void CloseTable(string path, Form parent = null)
        {
            EnviroRiskAddIn.Debug("Closing table " + path + " parent: " + parent.ToString());
            Semaphore s = new Semaphore(0, 1);

            if (parent == null)
                EnviroRiskAddIn.MBA.CallMapBasicSubroutine("CloseTAB", path);
            else
            {

                parent.BeginInvoke(new Action(() =>
                {
                    try
                    {
                        EnviroRiskAddIn.MBA.CallMapBasicSubroutine("CloseTAB", path);
                    }
                    finally
                    {
                        s.Release();
                    }
                }));
                s.WaitOne();
            }
        }

        public static void TabToRaster(string infile, string outfile)
        {
            PointTable table = new PointTable(infile);
            table.writeRaster(outfile);
        }


        public static bool RasterToCurves(string srcFilePath, string dstFilePath, double interval, bool fill = false, double levelMin=-50, double levelMax=12000)
        {
            bool polygons = fill;
            /*List<ManualLevel> niveles = new List<ManualLevel>();
            for(double i = levelMin; i < levelMax; i += interval)
            {
                niveles.Add(new ManualLevel(i));
            }
            */
            //                Levels = new ContourLevels(niveles),
            var options = new ContourOptions
            {
                InterpolationMethod = InterpolationMethod.Default,
                Levels = new ContourLevels(type: ContourLevelType.ConstantMinMax, spacing: interval),
                Styles = new ContourStyles(true),
                LevelRange = new ContourLevelRange(levelMin, levelMax)
            };
          
            var fieldBandFilter = new FieldBandFilter(0, 0);
            bool isGenerated = true;
            try
            {
                RasterAnalysis.Contour(srcFilePath, dstFilePath, 0, options, polygons, fieldBandFilter);
            }
            catch(Exception e)
            {
                EnviroRiskAddIn.Debug("Contour generation failed: " + e.ToString());
                isGenerated = false;
            }
            return isGenerated;
        }
    }
}
